local Boat = ...

-- State machine
StateMoving                     = 0
StateInitializingMoving         = 1
StateWaitingOnShipCollision     = 2
StateLaunching                  = 3
StateReturnInitializing         = 4
StateReturnMoving               = 5
StateReturnEndMoving            = 6
StateReturnWarningBehaviorTree  = 7
StateReturnEnd                  = 8

-- Exception
ExceptionOk               = 0
ExceptionNoWater          = 1
ExceptionCollision        = 2
ExceptionDoubleCollision  = 3

--------------------------------------------------------------------------
local HARBOUR_COMP_BOAT = {
	TypeName = HarbourPrefix .. "_COMP_BOAT",
	ParentType = _optionComp,
	Properties = {
 		{ Name = "MinimalSailors", Type = "integer", Default = 1},
 		{ Name = "RowingAnimation", Type = "boolean", Default = false},
 		{ Name = "SailingAnimation", Type = "boolean", Default = false},
 		{ Name = "FlagTerrain", Type = "integer", Default = 2 ^ OBJECT_FLAG.TERRAIN:toNumber() + 2 ^ OBJECT_FLAG.BUILDING_PART:toNumber() + 2 ^ OBJECT_FLAG.PLATFORM:toNumber()},
                { Name = "BoatSpeed", Type = "float", Default = 2.0 },
                { Name = "BoatRotationSpeed", Type = "float", Default = 0.2 },
                { Name = "BoatLength", Type = "float", Default = 10.0 },
                { Name = "BoatLengthRotationMarge", Type = "float", Default = 10.0 },
                { Name = "BoatSmoothRotation", Type = "boolean", Default = false },
                { Name = "RandomPath", Type = "boolean", Default = false },
                { Name = "MaxRandomDistanceMove", Type = "float", Default = 50.0 },
                { Name = "_onEnabledNumber", Type = "integer", Default = 0 },
                { Name = "FirstBoatInitialPosition", Type = "boolean", Default = true },
                { Name = "BoatInitialPosition", Type = "vec3f" },
                { Name = "BoatInitialOrientation", Type = "quaternion" },
                { Name = "IsInitDone", Type = "boolean", Default = false },
                { Name = "goBackToHarbourConditionInit", Type = "integer", Default = 80 },
                { Name = "_stateSequenceNumber", Type = "integer", Default = StateLaunching },
                { Name = "_timer", Type = "float", Default = 10 },
                { Name = "_errorCount", Type = "integer", Default = 0 },
                { Name = "_checkSailorStuckState", Type = "integer", Default = 0 },
                { Name = "_checkSailorStuckTimer", Type = "float", Default = 1 },
       	}
}

--------------------------------------------------------------------------
function HARBOUR_COMP_BOAT:initializeAssignSailor()
    self._assignedSailor = {}
    local bateau = self:getOwner()
    if (bateau == nil) then
        Boat:logError("HARBOUR_COMP_BOAT:initializeAssignSailor bateau nil")
        return
    end
    local boat = bateau:getParent()
    if (boat == nil) then
        Boat:logError("HARBOUR_COMP_BOAT:initializeAssignSailor boat nil")
        return
    end
    local comp = boat:getComponent("COMP_WORKPLACE")
    if (comp == nil) then
--        Boat:log("HARBOUR_COMP_BOAT:initializeAssignSailor comp_workplace nil")
        return
    end
    local workerList = comp.AssignedWorkerList
    if (workerList == nil) then
        Boat:logError("HARBOUR_COMP_BOAT:initializeAssignSailor workerList nil")
        return
    end
    if (#workerList <= 0) then
--        Boat:log("HARBOUR_COMP_BOAT:initializeAssignSailor workerList empty")
        return
    end
    if (#workerList > self.MinimalSailors) then
        Boat:log("HARBOUR_COMP_BOAT:initializeAssignSailor workerList out of bounds!!!! " .. #workerList)
        return
    end
    for i=1, #workerList do
--        Boat:log("HARBOUR_COMP_BOAT:workerList[" .. i .. "] = " .. tostring(workerList[i]))
        if (workerList[i] ~= nil) then
            local _sailor = workerList[i]:getOwner()
            if (_sailor == nil) then
               Boat:log("HARBOUR_COMP_BOAT:initializeAssignSailor _sailor nil")
            else
               table.insert(self._assignedSailor, _sailor)
            end
        end      
    end
--    Boat:log("HARBOUR_COMP_BOAT:initializeAssignSailor self._assignedSailor = " .. tostring(self._assignedSailor))
end
--------------------------------------------------------------------------
-- Called by the Behavior_tree
function HARBOUR_COMP_BOAT:registerAssignedSailor(_sailor)
--    Boat:log("COMP_BOAT:registerAssignedSailor()")
    if (_sailor == nil) or (_sailor.Name == nil) or (self.MinimalSailors == nil) then
        Boat:logError("registerAssignedSailor: error input")
        return -1
    end
    if self._assignedSailor == nil then
        Boat:log("registerAssignedSailor: _assignedSailor nil")
        return 1 -- by default we suppose the sailor will be assigned on next position
    end
    for i,alreadyRegisteredSailor in ipairs(self._assignedSailor) do
        if alreadyRegisteredSailor.Name == _sailor.Name then
--            Boat:log("registerAssignedSailor: _sailor already assigned found" .. tostring(_sailor))
            return i
        end
    end 
--    Boat:log("COMP_BOAT:assignedSailor no Registered" .. tostring(self._assignedSailor))
    return 1 -- by default we suppose the sailor will be assigned on next position
end

--------------------------------------------------------------------------
-- Called by the Behavior_tree
function HARBOUR_COMP_BOAT:checkBoatSailing(_name)
--        Boat:log("COMP_BOAT:checkBoatSailing()")
        if (self._stateSequenceNumber == nil) or (_name == nil) then
            Boat:logError("COMP_BOAT:checkBoatSailing: error input" .. tostring(self._assignedSailor))
            self._errorCount = self._errorCount + 1
            if self._errorCount > 3 then
                return BEHAVIOR_TREE_NODE_RESULT.FALSE -- to unlock the behavior tree in some cases
            end
        
        elseif (self._embarkedSailor == nil) or (self._assignedSailor == nil) then
            Boat:logError("COMP_BOAT:checkBoatSailing: error input" .. tostring(self._assignedSailor))
            self._errorCount = self._errorCount + 1
            if self._errorCount > 3 then
                return BEHAVIOR_TREE_NODE_RESULT.FALSE -- to unlock the behavior tree in some cases
            end
        elseif self._stateSequenceNumber == StateReturnWarningBehaviorTree then
            Boat:log("COMP_BOAT:checkBoatSailing: return FALSE" .. tostring(self._assignedSailor))
            for i,alreadyRegisteredSailor in ipairs(self._embarkedSailor) do
                if alreadyRegisteredSailor.Name == _name then
                    table.remove(self._embarkedSailor, i)
                    break
                end
            end
            if (#self._embarkedSailor <= 0) or (#self._assignedSailor <= 0) then
                self._isSailorsEmbarked = false
                self._stateSequenceNumber = StateReturnEnd
            end
            return BEHAVIOR_TREE_NODE_RESULT.FALSE

        elseif (#self._assignedSailor <= 0) then
            Boat:logError("COMP_BOAT:checkBoatSailing: error input self._assignedSailor")
            self._errorCount = self._errorCount + 1
            if self._errorCount > 3 then
                return BEHAVIOR_TREE_NODE_RESULT.FALSE -- to unlock the behavior tree in some cases
            end
        elseif self._isSailorsEmbarked then
            for i,alreadyRegisteredSailor in ipairs(self._embarkedSailor) do
                if alreadyRegisteredSailor.Name == _name then
                    return BEHAVIOR_TREE_NODE_RESULT.TRUE
                end
            end
            Boat:log("COMP_BOAT:checkBoatSailing: no assigned sailor found"  .. tostring(self._assignedSailor))
            return BEHAVIOR_TREE_NODE_RESULT.FALSE
        end
 --       Boat:log("COMP_BOAT:checkBoatSailing: return TRUE"  .. tostring(self._assignedSailor))
        return BEHAVIOR_TREE_NODE_RESULT.TRUE
end

--------------------------------------------------------------------------
function HARBOUR_COMP_BOAT:create()
    Boat:log("COMP_BOAT: create(): RandomPath = " .. tostring(self.RandomPath) .. tostring(self._assignedSailor))
    self._assignedSailor = {}
    self._initTimer = 10
end

--------------------------------------------------------------------------
function HARBOUR_COMP_BOAT:initMemorizationReturn()
        self.globalPositionPath = {}
        self.BoatSmoothRotationPath = {}
        self._initBoatRotationPath = {}
        self._timerPath = {}
        self._rotationYPath = {}
end

--------------------------------------------------------------------------
function HARBOUR_COMP_BOAT:init()
        Boat:log("COMP_BOAT: init: RandomPath = " .. tostring(self.RandomPath) .. tostring(self._assignedSailor))
--        self._onEnabledNumber = 0
        self._initTimer = 10
        self._rotationY = 0
        self.goBackToHarbourCondition = self.goBackToHarbourConditionInit
        self:initMemorizationReturn()
        self:initializeAssignSailor()
        self._embarkedSailor = {}
        self.globalPositionPathIndex = 0
        self.IsInitDone = true
        self._stateSequenceNumber = StateLaunching
        self._doubleCollision = false
        self._checkSailorStuckLocation = {}
        self._isSailorsEmbarked = false
        Boat:log("COMP_BOAT: init: end of method")
end

--------------------------------------------------------------------------
function HARBOUR_COMP_BOAT:checkSailorStuck(i, _sailor)
    if (self._checkSailorStuckState > self.MinimalSailors) then -- all sailors checked
        return
    end
    Boat:log("HARBOUR_COMP_BOAT:checkSailorStuck begin " .. tostring(_sailor))
    if (_sailor == nil) then
        Boat:log("HARBOUR_COMP_BOAT:checkSailorStuck sailor nil")
        return
    end
    local _elevationWater = computeElevation(self, _sailor:getGlobalPosition(), {OBJECT_FLAG.WATER})
    if (_elevationWater == nil) then
        Boat:log("HARBOUR_COMP_BOAT:checkSailorStuck _elevationWater nil")
        return
    end
    local _elevationGround = computeElevation(self, _sailor:getGlobalPosition(), {OBJECT_FLAG.TERRAIN, OBJECT_FLAG.PLATFORM, OBJECT_FLAG.BRIDGE})
    if (_elevationGround == nil) then
        Boat:log("HARBOUR_COMP_BOAT:checkSailorStuck _elevationGround nil")
        return
    end
    if (_elevationGround < _elevationWater) then
        Boat:log("HARBOUR_COMP_BOAT:checkSailorStuck sailor on water")
    else
        if (self._checkSailorStuckState == 0) then -- case of reset
            self._checkSailorStuckLocation = {}
            for i1,sailor1 in ipairs(self._assignedSailor) do
                Boat:log("HARBOUR_COMP_BOAT:checkSailorStuck initilization check stuck on ground  " .. tostring(sailor1) .. " " .. _elevationGround .. " " .. _elevationWater)
                table.insert(self._checkSailorStuckLocation, sailor1:getGlobalPosition())
                Boat:log("HARBOUR_COMP_BOAT:checkSailorStuck end initilization check stuck on ground")
            end 
            self._checkSailorStuckState = 1
        end       
        self._checkSailorStuckTimer = self._checkSailorStuckTimer - self:getLevel():getDeltaTime()
        if (self._checkSailorStuckLocation[i] == nil) then
            return -- I still don't know why this is necessary, but I have currently no time to investigate more
        end
        local deltaX = 	self._checkSailorStuckLocation[i][1] - _sailor:getGlobalPosition()[1] 
        local deltaZ = 	self._checkSailorStuckLocation[i][3] - _sailor:getGlobalPosition()[3] 
        local delta = deltaX*deltaX + deltaZ*deltaZ 
        if (delta > 1) then
            Boat:log("HARBOUR_COMP_BOAT:checkSailorStuck sailor on ground: ok")
            self._checkSailorStuckState = self._checkSailorStuckState + 1
            return --ok
        end
        if (self._checkSailorStuckTimer > 0) then
            return -- wait the check
        end
        Boat:log("HARBOUR_COMP_BOAT:checkSailorStuck sailor stuck on ground: not ok")
    end
    if ((self._guardingLocationComp == nil) or (self._guardingLocationComp[i] == nil)) then
        Boat:log("HARBOUR_COMP_BOAT:checkSailorStuck _guardingLocationComp not defined nil")
        return
    end
    Boat:log("HARBOUR_COMP_BOAT:checkSailorStuck begin try to fix sailor stuck")
    self._checkSailorStuckState = self._checkSailorStuckState + 1
    _sailor:setGlobalPosition(self._guardingLocationComp[i]:getOwner():getGlobalPosition()) --security to avoid the "drifts"
    _sailor:setGlobalOrientation(self._guardingLocationComp[i]:getOwner():getGlobalOrientation()) --security to avoid the "drifts"
    Boat:log("HARBOUR_COMP_BOAT:checkSailorStuck end try to fix sailor stuck")
    return -- fix done
end

--------------------------------------------------------------------------
function HARBOUR_COMP_BOAT:onDisabled()
--        Boat:log("COMP_BOAT:onDisabled()")
        self._onEnabledNumber = 0
end
--------------------------------------------------------------------------
function HARBOUR_COMP_BOAT:isSailorEmbarked()
        self._embarkedSailor = {}
        if self._assignedSailor == nil then
            Boat:logWarning("isSailorEmbarked: Sailor not assigned by Behavior_tree yet")
            Boat:log("isSailorEmbarked: RandomPath = " .. tostring(self.RandomPath) .. tostring(self._assignedSailor) .. self._onEnabledNumber)
            return false
        end
--        Boat:log("isSailorEmbarked: Check Minimal number of sailors " .. self.MinimalSailors)
        if #self._assignedSailor < self.MinimalSailors then
--            Boat:logWarning("isSailorEmbarked: Minimal number of sailors " .. self.MinimalSailors .. " not reached, actual = " .. #self._assignedSailor)
--            Boat:log("isSailorEmbarked: RandomPath = " .. tostring(self.RandomPath) .. tostring(self._assignedSailor) .. self._onEnabledNumber)
            return false
        end
        if self._onEnabledNumber < 1 then
            Boat:logWarning("isSailorEmbarked: Boat not enabled yet")
            Boat:log("isSailorEmbarked: RandomPath = " .. tostring(self.RandomPath) .. tostring(self._assignedSailor) .. self._onEnabledNumber)
 
            return false
        end
        self._guardingLocationComp = {}
        for i,sailor in ipairs(self._assignedSailor) do
            if (sailor ~= nil) then
                local _location = self:getOwner():findFirstObjectWithComponentDown(HarbourPrefix .. "_COMP_GUARDING_LOCATION" .. i)
                if _location == nil then
                    Boat:logError("isSailorEmbarked: " .. HarbourPrefix .. "_COMP_GUARDING_LOCATION" .. i .. " nil")
                    return false
                end
                table.insert(self._guardingLocationComp, _location)       

                local deltaX = 	_location:getOwner():getGlobalPosition()[1] - sailor:getGlobalPosition()[1] 
                local deltaZ = 	_location:getOwner():getGlobalPosition()[3] - sailor:getGlobalPosition()[3] 
                local delta = deltaX*deltaX + deltaZ*deltaZ 
                if delta <= 1 then
                    table.insert(self._embarkedSailor, sailor)
                else
--                Boat:log("isSailorEmbarked: Sailor " .. sailor.Name .. " not in Boat yet")
--                Boat:log("isSailorEmbarked: RandomPath = " .. tostring(self.RandomPath) .. tostring(self._assignedSailor) .. self._onEnabledNumber .. " delta = " .. delta)
                    self:checkSailorStuck(i, sailor) --security
                    if (self._checkSailorStuckState > self.MinimalSailors) then
                       self._checkSailorStuckState = self.MinimalSailors + 1
                    else
                       self._checkSailorStuckState = 1
                    end
                    return false
                end
            end
        end
        if self.SailingAnimation then
            local sailComponent = self:getOwner():findFirstObjectWithComponentDown(HarbourPrefix .. "_COMP_SAIL")
            if (sailComponent ~= nil) then
                sailComponent.windBlow = true
            end
        end
        if self.RowingAnimation then
            self:getOwner():forEachChildRecursive(
                function(child)
                    if (child.Name == "oars1") or (child.Name == "oars2") then
                        local oarComponent = child:findFirstObjectWithComponentDown(HarbourPrefix .. "_COMP_OAR")
                        if (oarComponent ~= nil) then
                            oarComponent.rowing = true
                        end
                    end
                end
            )  
        end
        self._isSailorsEmbarked = true
--        Boat:log("isSailorEmbarked: RandomPath = " .. tostring(self.RandomPath) .. tostring(self._assignedSailor))
        return true
end
--------------------------------------------------------------------------
function HARBOUR_COMP_BOAT:initializeEmbarkedSailor()
    self._embarkedSailor = {}
    for i,sailor in ipairs(self._assignedSailor) do
        table.insert(self._embarkedSailor, sailor)
    end
end
--------------------------------------------------------------------------
function HARBOUR_COMP_BOAT:launchShip()
--        Boat:log("COMP_BOAT: launchShip: FlagTerrain = " .. self.FlagTerrain .. " bit TERRAIN = " .. OBJECT_FLAG.TERRAIN:toNumber() .. " self._assignedSailor = " .. tostring(self._assignedSailor))
        -- memorize the position of the boat in the harbour
        if self.FirstBoatInitialPosition then
                self.BoatInitialPosition = self:getOwner():getGlobalPosition()
                self.BoatInitialOrientation = self:getOwner():getGlobalOrientation()
                self.FirstBoatInitialPosition = false
--                Boat:log("Memorize Boat Initial Position = " .. tostring(self.BoatInitialPosition) .. " list = " .. tostring(self.globalPositionPath))
        end
        self._rotationY = -math.pi/2
        self._initBoatRotationOffset = 0
        self.BoatLengthRotationMargeCurrent = self.BoatLengthRotationMarge
        self._isBoatLaunch = true
        -- memorize the elevation of the worker above the boat
        self.sailorDeltaY = self:getOwner():getGlobalPosition()[2] - self._assignedSailor[1]:getGlobalPosition()[2]
--        Boat:log("End of launchShip self._rotationY = " .. self._rotationY)
end
--------------------------------------------------------------------------
function HARBOUR_COMP_BOAT:onEnabled()
--        Boat:log("COMP_BOAT:onEnabled()")
        self._onEnabledNumber = 1
end
--------------------------------------------------------------------------
function HARBOUR_COMP_BOAT:chooseRandomSequence()
        if self._isBoatLaunch then
--            Boat:log("chooseRandomSequence sequence launch")
            self._initBoatRotation = 0 -- no rotation
            self._isBoatLaunch = false
            self.BoatSmoothRotation = true -- force Move only
        end
        if not self.BoatSmoothRotation then
--            Boat:log("chooseRandomSequence rotate and move initialization")
            self._initBoatRotation = self._initBoatRotationOffset + (math.random() - 0.5) * math.pi/6
            if self._initBoatRotation < 0 then
                self._initBoatRotation = self._initBoatRotation + 2*math.pi
            end
            if self._initBoatRotation > math.pi then
                self._initBoatRotation = self._initBoatRotation - 2*math.pi
            end
            self._timer = math.abs(self._initBoatRotation) / self.BoatRotationSpeed
            self._initBoatRotationOffset = 0
            self.BoatSmoothRotation = true
        else
--            Boat:log("chooseRandomSequence Move only initialization")
            self.BoatLengthRotationMargeCurrent = self.BoatLengthRotationMarge
            self._timer = math.random() * self.MaxRandomDistanceMove / self.BoatSpeed -- straight move
            self.BoatSmoothRotation = false
        end
end

--------------------------------------------------------------------------
-- the object is supposed to be vertical aligned
function HARBOUR_COMP_BOAT:getSinusFromQuaternion(_q)
    return 2 * _q[4] * _q[2]
end

function HARBOUR_COMP_BOAT:getCosinusFromQuaternion(_q)
    return _q[4] * _q[4] - _q[2] * _q[2]
end

function HARBOUR_COMP_BOAT:getAngleFromQuaternion(_q)
        local sinp = self:getSinusFromQuaternion(_q)
        local cosp = self:getCosinusFromQuaternion(_q)
        local normep = cosp*cosp + sinp*sinp
        if (normep < 0.99) or (normep > 1.01)  then
            Boat:logError("getAngleFromQuaternion: getGlobalOrientation() error: " .. tostring(_q))
            return nil
        end
        if sinp < -1 then -- security, should never occur
            sinp = -1
        elseif sinp > 1 then
            sinp = 1
        end
        if cosp < -1 then -- security, should never occur
            cosp = -1
        elseif cosp > 1 then
            cosp = 1
        end
        local _angleCos = math.acos(cosp)
        if sinp < 0 then
            _angleCos = - _angleCos
        end
        if _angleCos < 0 then
           _angleCos = _angleCos + 2*math.pi
        end
        local _angleSin = math.asin(sinp)
        if cosp < 0 then
            _angleSin = math.pi - _angleSin
        end
        if _angleSin < 0 then
           _angleSin = _angleSin + 2*math.pi
        end
        if (_angleCos*_angleCos - _angleSin*_angleSin) > 0.01 then
            Boat:logError("getAngleFromQuaternion: updateBoatNextMove accuracy error")
        end
--        Boat:log("getAngleFromQuaternion : sinp = " .. sinp .. " cosp = " .. cosp .. " asinp = " .. (_angleSin/math.pi) .. " * pi" .. " acosp = " .. (_angleCos/math.pi) .. " * pi" .. " return = " .. (((_angleCos + _angleSin) / 2)/math.pi) .. " * pi")
        return (_angleCos + _angleSin) / 2
end

function HARBOUR_COMP_BOAT:getRotationYFromTarget2(_nextPosition, _initialBoatOrientation)
--    Boat:log("getRotationYFromTarget _nextPosition = " .. tostring(_nextPosition))
    local DeltaX = _nextPosition[1] - self:getOwner():getGlobalPosition()[1]
    local DeltaZ = _nextPosition[3] - self:getOwner():getGlobalPosition()[3]
    local Distance = (DeltaX*DeltaX + DeltaZ*DeltaZ) ^ 0.5
    if Distance <= 0 then
        return 0
    end
    local cosp = DeltaX / Distance
    local sinp = DeltaZ / Distance
    if cosp < -1 then -- security, should never occur
        cosp = -1
    elseif cosp > 1 then
        cosp = 1
    end
    local _angleCos = math.acos(cosp)
    if sinp < 0 then
        _angleCos = - _angleCos
    end
    _angleCos = math.pi/2 - _angleCos
    local _angle = _angleCos -self:getAngleFromQuaternion(_initialBoatOrientation)
--    Boat:log("getRotationYFromTarget2 : return = " .. (_angle/math.pi) .. " * pi ; current Angle : return = " .. (_angleCos/math.pi) .. " * pi")
    return _angle
end

--------------------------------------------------------------------------
function HARBOUR_COMP_BOAT:refreshSailorsPosition(_delta)
    if self._assignedSailor == nil then
        Boat:logError("COMP_BOAT:refreshSailorsPosition error input")
        return
    end
    if #self._assignedSailor <= 0 then
        Boat:logError("COMP_BOAT:refreshSailorsPosition error input")
        return
    end
    for i,sailor in ipairs(self._assignedSailor) do
        sailor:setGlobalPosition(self._guardingLocationComp[i]:getOwner():getGlobalPosition()) --security to avoid the "drifts"
        sailor:rotateY(_delta)
    end
end

--------------------------------------------------------------------------
function HARBOUR_COMP_BOAT:refreshBoatRotation(_rotationY, _delta)
    self:getOwner():setRotationY(_rotationY + math.pi/2) -- +pi/2 to take into account the rotation due to the boat model
    self:refreshSailorsPosition(_delta)
end

--------------------------------------------------------------------------
function HARBOUR_COMP_BOAT:moveBoatOnDt(_dt)
 --       Boat:log("moveBoatOnDt")
        -- rotation          
        if self.BoatSmoothRotation then
            if self._initBoatRotation < 0 then
                self._rotationY = self._rotationY - _dt*self.BoatRotationSpeed               
                self:refreshBoatRotation(self._rotationY, - _dt*self.BoatRotationSpeed)
            else
                self._rotationY = self._rotationY + _dt*self.BoatRotationSpeed               
                self:refreshBoatRotation(self._rotationY, _dt*self.BoatRotationSpeed)
            end
        end
    -- Raycast from the last current position to the next position + BoatLength + BoatLengthRotationMargeCurrent
    -- only on objects with FlagTerrain
        local raycastResult = {}
        local q = self:getOwner():getGlobalOrientation()
        local sinp = self:getSinusFromQuaternion(q)
        local cosp = self:getCosinusFromQuaternion(q)
        -- take into account the orientation of the boat in the model
        local BoatExtremity = { self:getOwner():getGlobalPosition()[1] - cosp * (self.BoatLength+self.BoatLengthRotationMargeCurrent), 
                                  self:getOwner():getGlobalPosition()[2],
                                  self:getOwner():getGlobalPosition()[3] + sinp * (self.BoatLength+self.BoatLengthRotationMargeCurrent)}
        if self:getLevel():rayCast(self:getOwner():getGlobalPosition(), BoatExtremity, raycastResult, self.FlagTerrain) then
--            Boat:log("collision with TERRAIN raycastResult = RigidBody = " .. tostring(raycastResult["RigidBody"]) .. " Position = " .. tostring(raycastResult["Position"]) .. " Normal = " .. tostring(raycastResult["Normal"]))
            local pvector = (cosp*raycastResult["Normal"][3] + sinp*raycastResult["Normal"][1])*self.BoatLengthRotationMarge
            if pvector == 0 then  -- security, but should quasi never occur
                Boat:log("moveBoatOnDt collision with TERRAIN pvector == 0")
                pvector = self.BoatLengthRotationMarge
            end
            local reflection = { self:getOwner():getGlobalPosition()[1] - raycastResult["Normal"][3]*pvector, self:getOwner():getGlobalPosition()[2], self:getOwner():getGlobalPosition()[3] + raycastResult["Normal"][1]*pvector }
            self._initBoatRotationOffset = self:getRotationYFromTarget2(reflection, self.BoatInitialOrientation) - self._rotationY
            self.BoatLengthRotationMargeCurrent = 0 -- to let the time for the ship to turn without reaching again the terrain collision condition
            if self._doubleCollision then
--                Boat:log("moveBoatOnDt ExceptionDoubleCollision")
                return ExceptionDoubleCollision
            end
            self._doubleCollision =true
            return ExceptionCollision -- TERRAIN Reached
        end
        self._doubleCollision =false

    -- compute the next water elevation: p is the rotation since the creation of the boat. The orientation of the model is taken into account    
        local BoatDestination = { self:getOwner():getGlobalPosition()[1] - cosp * _dt * self.BoatSpeed, 
                                     self:getOwner():getGlobalPosition()[2],
                                     self:getOwner():getGlobalPosition()[3] + sinp * _dt * self.BoatSpeed}
        local _elevation = computeElevation(self, BoatDestination, {OBJECT_FLAG.WATER})
        if _elevation == nil then
            return ExceptionNoWater
        end
    -- moves 
        self:getOwner():move({ -_dt*self.BoatSpeed, _elevation-BoatDestination[2], 0 })
        if self._assignedSailor == nil then
            Boat:log("moveBoatOnDt self._assignedSailor nil")
            return ExceptionOk -- ok    
        end    
        if #self._assignedSailor <= 0 then
            Boat:log("moveBoatOnDt self._assignedSailor empty")
            return ExceptionOk -- ok    
        end
        for i,sailor in ipairs(self._assignedSailor) do
            sailor:setGlobalPosition(self._guardingLocationComp[i]:getOwner():getGlobalPosition()) --security to avoid the "drifts"
        end
--        Boat:log("end moveBoatOnDt " .. tostring(self._assignedSailor))
        return ExceptionOk -- ok    
end

--------------------------------------------------------------------------
function HARBOUR_COMP_BOAT:update()
--    Boat:log("update()")
    if not self.IsInitDone then -- security
        return
    end
    local dt = self:getLevel():getDeltaTime()

    self:initializeAssignSailor()
--    Boat:log("update() " .. tostring(self._assignedSailor))
-- sequences for casual moves
    if self._stateSequenceNumber == StateMoving then      
        -- moving sequence
--        Boat:log("StateMoving self._timer = " .. self._timer .. tostring(self._assignedSailor))
        self._timer = self._timer - dt 
        local cret = self:moveBoatOnDt(dt)
--        Boat:log("return from self:moveBoatOnDt cret = " .. cret .. tostring(self._assignedSailor))
        if cret == ExceptionOk then
            if self._timer < 0 then
--                Boat:log("StateInitializingMoving" .. tostring(self._assignedSailor))
                self._stateSequenceNumber = StateInitializingMoving
            end
        elseif (cret == ExceptionCollision) or (cret == ExceptionDoubleCollision) then -- case of collision
--                Boat:log("COMP_BOAT:update: StateMoving ExceptionCollision" .. tostring(self._assignedSailor))
            self.BoatSmoothRotation = false -- to initialize the first rotation
            self._timerPath[#self._timerPath] = self._timerPath[#self._timerPath] - self._timer + dt  -- memorize that this path is shorter  
--            Boat:log("StateInitializingMoving" .. tostring(self._assignedSailor))
            self._stateSequenceNumber = StateInitializingMoving
        elseif cret == ExceptionNoWater then --error detected by checkWaterElevation
--                Boat:log("COMP_BOAT: update: StateMoving ExceptionNoWater" .. tostring(self._assignedSailor))
            self.BoatLengthRotationMargeCurrent = 0 -- to let the time for the ship to turn without reaching again the terrain collision condition
            self.globalPositionPathIndex = #self.globalPositionPath
            self._isBoatLaunch = true -- force the return without smoothing
--            Boat:log("StateReturnInitializing" .. tostring(self._assignedSailor))
            self._stateSequenceNumber = StateReturnInitializing
        end
--        Boat:log("update() end StateMoving")
        return

    elseif self._stateSequenceNumber == StateInitializingMoving then            
        -- orientation setting
--        Boat:log("StateInitializingMoving self._timer = " .. self._timer)
        if self.RandomPath then
            -- count down to the next return to harbour        
            self.goBackToHarbourCondition = self.goBackToHarbourCondition - 1
            if self.goBackToHarbourCondition <= 0 then
--                Boat:log("COMP_BOAT: update: StateInitializingMoving self.goBackToHarbourCondition <= 0")
                self.BoatLengthRotationMargeCurrent = 0 -- to let the time for the ship to turn without reaching again the terrain collision condition
                self.globalPositionPathIndex = #self.globalPositionPath
                self._isBoatLaunch = true -- force the return without smoothing
--                Boat:log("StateReturnInitializing" .. tostring(self._assignedSailor))
                self._stateSequenceNumber = StateReturnInitializing
                return                            
            end
            -- choose a random destination
            self:chooseRandomSequence()
            -- memorize the last position , and the parameters of the next rotation/move
            table.insert(self.globalPositionPath, self:getOwner():getGlobalPosition())
            table.insert(self.BoatSmoothRotationPath, self.BoatSmoothRotation)
            table.insert(self._initBoatRotationPath, self._initBoatRotation)
            table.insert(self._timerPath, self._timer)
            table.insert(self._rotationYPath, self._rotationY)
--            Boat:log("Memorisation Sequence (" .. #self.globalPositionPath .. ","  .. #self.BoatSmoothRotationPath .. "," .. #self._initBoatRotationPath .. "," .. #self._timerPath .. ") _timer = " .. self._timer .. " _rotationY = " .. self._rotationY)
--            Boat:log("StateMoving" .. tostring(self._assignedSailor))
            self._stateSequenceNumber = StateMoving
        end
        return

-- Sequences for the boat returning to harbour
    elseif self._stateSequenceNumber == StateReturnInitializing then
--        Boat:log("COMP_BOAT:update: StateReturnInitializing self.globalPositionPathIndex = " .. self.globalPositionPathIndex)
        if self.globalPositionPathIndex <= 0 then
--            Boat:log("COMP_BOAT:update: StateReturnInitializing back to harbour self.globalPositionPathIndex = " .. self.globalPositionPathIndex)

            self:getOwner():setGlobalTransform(self.BoatInitialPosition, self.BoatInitialOrientation)
            if (self._assignedSailor ~= nil) and (#self._assignedSailor > 0) then
                for i,sailor in ipairs(self._assignedSailor) do
                    sailor:setGlobalPosition(self.BoatInitialPosition)
                end
            end
            self._rotationY = - self:getAngleFromQuaternion(self.BoatInitialOrientation)
            if self.SailingAnimation then
                local sailComponent = self:getOwner():findFirstObjectWithComponentDown(HarbourPrefix .. "_COMP_SAIL")
                if (sailComponent ~= nil) then
                    sailComponent.windBlow = false
                end
            end
            if self.RowingAnimation then
                self:getOwner():forEachChildRecursive(
                    function(child)
                        if (child.Name == "oars1") or (child.Name == "oars2") then
                            local oarComponent = child:findFirstObjectWithComponentDown(HarbourPrefix .. "_COMP_OAR")
                            if (oarComponent ~= nil) then
                                oarComponent.rowing = false
                            end
                        end
                    end
                )  
            end
            
            self:initializeEmbarkedSailor()
--            Boat:log("HARBOUR_COMP_BOAT:update self._embarkedSailor = " .. #self._embarkedSailor)
            if (#self._embarkedSailor > 0) then
                self._stateSequenceNumber =  StateReturnWarningBehaviorTree
            else
                self._stateSequenceNumber =  StateReturnEnd
            end
            return
        end 
--        Boat:log("COMP_BOAT:update: StateReturnInitializing self.globalPositionPathIndex = " .. self.globalPositionPathIndex .. " self._timerPath[self.globalPositionPathIndex] = " .. self._timerPath[self.globalPositionPathIndex])
        -- first move for return
        if self._isBoatLaunch then
            self._rotationY = self._rotationY + math.pi
            self:refreshBoatRotation(self._rotationY, math.pi)
            self._isBoatLaunch = false
        end  
        self.BoatSmoothRotation = self.BoatSmoothRotationPath[self.globalPositionPathIndex]
        self._initBoatRotation = - self._initBoatRotationPath[self.globalPositionPathIndex]
        self._initTimer = self._timerPath[self.globalPositionPathIndex]
           
        if self._initTimer >= 0 then
            self._stateSequenceNumber = StateReturnMoving
        else
--            Boat:log("StateReturnInitializing: self._initTimer = " .. self._initTimer .. " self.globalPositionPathIndex = " .. self.globalPositionPathIndex)
            self.globalPositionPathIndex = self.globalPositionPathIndex - 1
        end
        return

    elseif self._stateSequenceNumber == StateReturnWarningBehaviorTree then -- Free the remaining behavior tree
            if (#self._embarkedSailor <= 0) or (#self._assignedSailor <= 0) then
                self._stateSequenceNumber =  StateReturnEnd -- free COMP_BOAT
            end
            return

    elseif self._stateSequenceNumber == StateReturnMoving then -- rotation and move
--        Boat:log("COMP_BOAT:update: StateReturnMoving: self._rotationY = " .. self._rotationY .. " self._initTimer = " .. self._initTimer .. " self.globalPositionPathIndex = " .. self.globalPositionPathIndex)
        self._initTimer = self._initTimer - dt
        -- move
        local exception = self:moveBoatOnDt(dt)
        if exception == ExceptionNoWater then
--            Boat:log("COMP_BOAT:update: StateReturnMoving Move back to Position: No water self.globalPositionPathIndex = " .. self.globalPositionPathIndex .. " self._timerPath[self.globalPositionPathIndex] = " .. self._timerPath[self.globalPositionPathIndex])
            self._stateSequenceNumber = StateReturnEndMoving
            return
        end
        if exception == ExceptionCollision then
--            Boat:logError("COMP_BOAT:update: StateReturnMoving: Terrain collision self.globalPositionPathIndex = " .. self.globalPositionPathIndex .. " self._timerPath[self.globalPositionPathIndex] = " .. self._timerPath[self.globalPositionPathIndex])
            self._stateSequenceNumber = StateReturnEndMoving
            return
        end
        if exception == ExceptionDoubleCollision then
--            Boat:logError("COMP_BOAT:update: StateReturnMoving: Terrain double collision self.globalPositionPathIndex = " .. self.globalPositionPathIndex .. " self._timerPath[self.globalPositionPathIndex] = " .. self._timerPath[self.globalPositionPathIndex])
            self.globalPositionPathIndex = self.globalPositionPathIndex - 1
            self._stateSequenceNumber = StateReturnEndMoving
            return
        end
        if self._initTimer < 0 then
--            Boat:log("COMP_BOAT:update: StateReturnMoving: self._initTimer < 0 self.globalPositionPathIndex = " .. self.globalPositionPathIndex .. " self._timerPath[self.globalPositionPathIndex] = " .. self._timerPath[self.globalPositionPathIndex])
            self._stateSequenceNumber = StateReturnEndMoving
        end
        return

    elseif self._stateSequenceNumber == StateReturnEndMoving then -- next step
--        Boat:log("COMP_BOAT:update: StateReturnEndMoving self.globalPositionPathIndex = " .. self.globalPositionPathIndex)
        if self.globalPositionPathIndex > 0 then
--          Boat:log("COMP_BOAT:update: StateReturnEndMoving self.globalPositionPathIndex = " .. self.globalPositionPathIndex .. " self._timerPath[self.globalPositionPathIndex] = " .. self._timerPath[self.globalPositionPathIndex])
            -- fixes the drift
            self:getOwner():setGlobalPosition(self.globalPositionPath[self.globalPositionPathIndex])
            local _delta = self._rotationYPath[self.globalPositionPathIndex] + math.pi - self._rotationY
            self._rotationY = self._rotationYPath[self.globalPositionPathIndex] + math.pi
            -- the refreshBoatRotation also update the sailor position
            self:refreshBoatRotation(self._rotationY, _delta) -- needed in some rare case of coast configuration
            -- new back path           
            self.globalPositionPathIndex = self.globalPositionPathIndex - 1
        end 
        self._stateSequenceNumber = StateReturnInitializing
        return

    elseif self._stateSequenceNumber == StateReturnEnd then
--        Boat:log("COMP_BOAT:update: StateReturnEnd self.globalPositionPathIndex = " .. self.globalPositionPathIndex)
        self._initTimer = 30 -- wait 
        self._stateSequenceNumber = StateLaunching -- wait sailor embarked
        return

-- sequences for launching boat (first boat setting after a sailor assignment, after a return to harbour, or after a reloaded game)
    elseif self._stateSequenceNumber == StateLaunching then
        -- case of reload
        if self._initTimer >= 0 then
            self._initTimer = self._initTimer - dt
            return
        end
--        Boat:log("StateLaunching wait for init done")
        if not self.IsInitDone then
            self._initTimer = 10
            return
        end
--        Boat:log("COMP_BOAT: update: StateLaunching  _onEnabledNumber = " .. self._onEnabledNumber)
        if self._onEnabledNumber == 0 then
            self._initTimer = 10
            return
        end
--        Boat:log("StateLaunching _onEnabledNumber > 0")

        if not self:isSailorEmbarked() then
--            Boat:log("StateLaunching isSailorEmbarked() false" .. tostring(self._assignedSailor))
            self._initTimer = 10
            return
        end
--        Boat:log("StateLaunching sailor embarked" .. tostring(self._assignedSailor))
        self:launchShip()
        self:initMemorizationReturn()
        self.goBackToHarbourCondition =  self.goBackToHarbourConditionInit -- no immediate return to harbour
        self.BoatSmoothRotation = false -- to initialize the first rotation
        self._stateSequenceNumber = StateInitializingMoving
    end
end

--------------------------------------------------------------------------
Boat:registerClass(HARBOUR_COMP_BOAT)
